@@ -12,7 +12,8 @@ except ImportError:
 
 EXT_MODULES = ["lxml.etree", "lxml.objectify"]
 
-PACKAGE_PATH = "src/lxml/"
+PACKAGE_PATH = "src%slxml%s" % (os.path.sep, os.path.sep)
+INCLUDE_PACKAGE_PATH = PACKAGE_PATH + 'includes'
 
 if sys.version_info[0] >= 3:
     _system_encoding = sys.getdefaultencoding()
@@ -55,26 +56,36 @@ def ext_modules(static_include_dirs, static_library_dirs,
                 libxslt_version=OPTION_LIBXSLT_VERSION,
                 multicore=OPTION_MULTICORE)
 
-    if CYTHON_INSTALLED:
-        source_extension = ".pyx"
-        print("Building with Cython %s." % Cython.Compiler.Version.version)
-
-        from Cython.Compiler import Options
-        Options.generate_cleanup_code = 2
-    else:
-        source_extension = ".c"
-        if not os.path.exists(PACKAGE_PATH + 'lxml.etree.c'):
-            print ("WARNING: Trying to build without Cython, but pre-generated "
-                   "'%slxml.etree.c' does not seem to be available." % PACKAGE_PATH)
-        else:
-            print ("Building without Cython.")
-
     if OPTION_WITHOUT_OBJECTIFY:
         modules = [ entry for entry in EXT_MODULES
                     if 'objectify' not in entry ]
     else:
         modules = EXT_MODULES
 
+    c_files_exist = [ os.path.exists('%s%s.c' % (PACKAGE_PATH, module)) for module in modules ]
+
+    if CYTHON_INSTALLED and (OPTION_WITH_CYTHON or False in c_files_exist):
+        source_extension = ".pyx"
+        print("Building with Cython %s." % Cython.Compiler.Version.version)
+        # generate module cleanup code
+        from Cython.Compiler import Options
+        Options.generate_cleanup_code = 3
+    elif not OPTION_WITHOUT_CYTHON and False in c_files_exist:
+        for exists, module in zip(c_files_exist, modules):
+            if not exists:
+                raise RuntimeError(
+                    "ERROR: Trying to build without Cython, but pre-generated '%s%s.c' "
+                    "is not available (pass --without-cython to ignore this error)." % (
+                        PACKAGE_PATH, module))
+    else:
+        if False in c_files_exist:
+            for exists, module in zip(c_files_exist, modules):
+                if not exists:
+                    print("WARNING: Trying to build without Cython, but pre-generated "
+                          "'%s%s.c' is not available." % (PACKAGE_PATH, module))
+        source_extension = ".c"
+        print("Building without Cython.")
+
     lib_versions = get_library_versions()
     if lib_versions[0]:
         print("Using build configuration of libxml2 %s and libxslt %s" % 
@@ -89,6 +100,8 @@ def ext_modules(static_include_dirs, static_library_dirs,
     _define_macros = define_macros()
     _libraries = libraries()
 
+    _include_dirs.append(os.path.join(get_base_dir(), INCLUDE_PACKAGE_PATH))
+
     if _library_dirs:
         message = "Building against libxml2/libxslt in "
         if len(_library_dirs) > 1:
@@ -104,42 +117,50 @@ def ext_modules(static_include_dirs, static_library_dirs,
     else:
         runtime_library_dirs = []
 
-    if OPTION_SHOW_WARNINGS:
-        if CYTHON_INSTALLED:
-            from Cython.Compiler import Errors
-            Errors.LEVEL = 0
-    else:
-        _cflags = ['-w'] + _cflags
+    if CYTHON_INSTALLED and OPTION_SHOW_WARNINGS:
+        from Cython.Compiler import Errors
+        Errors.LEVEL = 0
 
     result = []
     for module in modules:
         main_module_source = PACKAGE_PATH + module + source_extension
         result.append(
             Extension(
-            module,
-            sources = [main_module_source],
-            depends = find_dependencies(module),
-            extra_compile_args = _cflags,
-            extra_objects = static_binaries,
-            define_macros = _define_macros,
-            include_dirs = _include_dirs,
-            library_dirs = _library_dirs,
-            runtime_library_dirs = runtime_library_dirs,
-            libraries = _libraries,
+                module,
+                sources = [main_module_source],
+                depends = find_dependencies(module),
+                extra_compile_args = _cflags,
+                extra_objects = static_binaries,
+                define_macros = _define_macros,
+                include_dirs = _include_dirs,
+                library_dirs = _library_dirs,
+                runtime_library_dirs = runtime_library_dirs,
+                libraries = _libraries,
             ))
+    if CYTHON_INSTALLED and OPTION_WITH_CYTHON_GDB:
+        for ext in result:
+            ext.cython_gdb = True
+
+    if CYTHON_INSTALLED and source_extension == '.pyx':
+        # build .c files right now and convert Extension() objects
+        from Cython.Build import cythonize
+        result = cythonize(result)
+
     return result
 
 def find_dependencies(module):
     if not CYTHON_INSTALLED:
         return []
-    package_dir = os.path.join(get_base_dir(), PACKAGE_PATH)
-    files = os.listdir(package_dir)
-    pxd_files = [ os.path.join(PACKAGE_PATH, filename) for filename in files
+    base_dir = get_base_dir()
+    package_dir = os.path.join(base_dir, PACKAGE_PATH)
+    includes_dir = os.path.join(base_dir, INCLUDE_PACKAGE_PATH)
+    pxd_files = [ os.path.join(includes_dir, filename)
+                  for filename in os.listdir(includes_dir)
                   if filename.endswith('.pxd') ]
 
     if 'etree' in module:
         pxi_files = [ os.path.join(PACKAGE_PATH, filename)
-                      for filename in files
+                      for filename in os.listdir(package_dir)
                       if filename.endswith('.pxi')
                       and 'objectpath' not in filename ]
         pxd_files = [ filename for filename in pxd_files
@@ -229,6 +250,8 @@ def define_macros():
         macros.append(('WITHOUT_THREADING', None))
     if OPTION_WITH_REFNANNY:
         macros.append(('CYTHON_REFNANNY', None))
+    if OPTION_WITH_UNICODE_STRINGS:
+        macros.append(('LXML_UNICODE_STRINGS', '1'))
     return macros
 
 _ERROR_PRINTED = False
@@ -338,9 +361,12 @@ def option_value(name):
 staticbuild = bool(os.environ.get('STATICBUILD', ''))
 # pick up any commandline options and/or env variables
 OPTION_WITHOUT_OBJECTIFY = has_option('without-objectify')
+OPTION_WITH_UNICODE_STRINGS = has_option('with-unicode-strings')
 OPTION_WITHOUT_ASSERT = has_option('without-assert')
 OPTION_WITHOUT_THREADING = has_option('without-threading')
 OPTION_WITHOUT_CYTHON = has_option('without-cython')
+OPTION_WITH_CYTHON = has_option('with-cython')
+OPTION_WITH_CYTHON_GDB = has_option('cython-gdb')
 OPTION_WITH_REFNANNY = has_option('with-refnanny')
 if OPTION_WITHOUT_CYTHON:
     CYTHON_INSTALLED = False
